@@ -7,6 +7,8 @@ import os
 import shutil
 from copy import deepcopy as dcopy
 
+import numpy as np
+
 import torch
 import torch.optim as optim
 import torch.utils.data as D
@@ -18,6 +20,7 @@ import tools.modules
 # from . import validation as V
 from . import utils as U
 from . import eval as E
+from . import model as M
 
 # __all__ = [
 # 'Trainer'
@@ -28,7 +31,7 @@ log = logging.getLogger(__name__)
 
 # %%
 class Trainer:
-    def __init__(self, network, train_dataset=None, cfg_train={}, criterion=None, network_dir=None, node_dir=None, cfg_val=None, val_dataset=None, val_metrics=None, name='default', verbose=True, amp=False, reproduce=True):
+    def __init__(self, network, train_dataset=None, cfg_train={}, criterion=None, network_dir=None, model_dir=None, cfg_val=None, val_dataset=None, val_metrics=None, name='default', verbose=True, amp=False, reproduce=True):
         '''
         Things Trainer do:
         - train network with early stopping, given hyperparameters
@@ -59,7 +62,7 @@ class Trainer:
         Recommended way of making hs.node.Node object is like the following:
 
             kwargs = {'network': Network(), 'train_dataset': train_dataset, 'validation': None, 'cfg_train': cfg.train,
-                    'criterion': criterion, 'network_dir': path['network'], 'node_dir': path['node'], 'verbose': True, 'amp': True}
+                    'criterion': criterion, 'network_dir': path['network'], 'model_dir': path['node'], 'verbose': True, 'amp': True}
             node = hs.node.Node(**kwargs)
         '''
         # Store configurations
@@ -68,15 +71,16 @@ class Trainer:
         self.cfg_train = dcopy(dict(cfg_train))
         self.criterion = criterion
         self.network_dir = network_dir
-        self.node_dir = node_dir
+        self.model_dir = model_dir
 
         self.cfg_val = dcopy(dict(cfg_val)) if cfg_val is not None else {}
         self.val_dataset = val_dataset
         self.val_metrics = val_metrics
         self.earlystopper = EarlyStopper(increase_better=cfg_val['increase_better'], patience=cfg_val['patience']) if self.val_dataset is not None else None
         if 'target_dataset' in self.cfg_val:
-            assert self.cfg_val['target_dataset'] in self.val_dataset, f'[Validation setup] target_dataset: {self.cfg_val["target_dataset"]} not found in val_dataset: {self.val_dataset.keys()}'
-            assert ~(self.val_dataset is not None ^ self.val_metrics is not None), f'val_dataset and val_metrics must be given together or neither. val_dataset: {self.val_dataset}, val_metrics: {self.val_metrics}'
+            if type(self.val_dataset)==dict:
+                assert self.cfg_val['target_dataset'] in self.val_dataset, f'[Validation setup] target_dataset: {self.cfg_val["target_dataset"]} not found in val_dataset: {self.val_dataset.keys()}'
+            assert ~((self.val_dataset is not None) ^ (self.val_metrics is not None)), f'val_dataset and val_metrics must be given together or neither. val_dataset: {self.val_dataset}, val_metrics: {self.val_metrics}'
 
         self.name = name
         self.verbose = verbose
@@ -88,9 +92,9 @@ class Trainer:
             if os.path.exists(self.network_dir): log.warning(f'path: {self.network_dir} exists. Be careful')
             os.makedirs(self.network_dir, exist_ok=True)
 
-        if self.node_dir is not None:
-            if os.path.exists(self.node_dir): log.warning(f'path: {self.node_dir} exists. Be careful')
-            os.makedirs(self.node_dir, exist_ok=True)
+        if self.model_dir is not None:
+            if os.path.exists(self.model_dir): log.warning(f'path: {self.model_dir} exists. Be careful')
+            os.makedirs(self.model_dir, exist_ok=True)
 
         self.train_meter = tools.modules.AverageMeter() # Tracks loss per epoch
         self.loss_tracker = tools.modules.ValueTracker() # Tracks loss over all training
@@ -111,7 +115,7 @@ class Trainer:
 
     def set_model(self):
         '''Set miscellaneous variables'''
-        self.model = BaseModel(self.network, amp=self.amp)
+        self.model = M.BaseModel(self.network, amp=self.amp)
         self.misc = tools.TDict()
         if self.train_dataset is not None:
             if hasattr(self.train_dataset, 'get_f'): # get_f equal to preprocessing pipeline before feeding to network
@@ -125,24 +129,36 @@ class Trainer:
             if type(self.val_dataset)==dict:
                 scores = {}
                 for name, dataset in self.val_dataset.items():
-                    d_results = E.test_model(self.model, dataset, batch_size=self.cfg_train['batch_size'], amp=self.amp)
-                    score = evaluate(d_results, self.val_metrics)
+                    d_results = E.test_model(self.model, dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
+                    if any([np.isnan(v).any() for v in d_results.values()]):
+                        log.warning(f'[Node: {self.name}][validate] NaN found in d_results. Skipping validation...')
+                        score = {self.cfg_val['criterion']: self.earlystopper.best_score}
+                    else:
+                        score = E.evaluate(d_results, self.val_metrics)
                     scores[name] = score
                     self.print(f'[dataset_name: {name}] Validation Score: {score}')
                 score = scores[self.cfg_val['target_dataset']]
             else:
-                d_results = E.test_model(self.model, self.val_dataset, batch_size=self.cfg_train['batch_size'], amp=self.amp)
-                score = evaluate(d_results, self.val_metrics)
+                d_results = E.test_model(self.model, self.val_dataset, batch_size=self.cfg_val['batch_size'], amp=self.amp)
+                if any([np.isnan(v).any() for v in d_results.values()]):
+                    log.warning(f'[Node: {self.name}][validate] NaN found in d_results. Skipping validation...')
+                    score = {self.cfg_val['criterion']: self.earlystopper.best_score}
+                else:
+                    score = E.evaluate(d_results, self.val_metrics)
                 self.print(f'Validation Score: {score}')
             
             if self.earlystopper is not None:
                 if type(score) is dict:
                     score_ = score[self.cfg_val['criterion']]
-
+                else:
+                    score_ = score
+                    
                 patience_end = self.earlystopper.step(self, score_)
                 if patience_end:
                     self.print('patience met, flushing earlystopper history')
                     self.earlystopper.history.reset()
+                    
+        torch.cuda.empty_cache()
         return patience_end
 
     def generate_loader(self):
@@ -249,7 +265,7 @@ class Trainer:
 
             self.iter += 1
             loss = self._update(data)
-            self.print(f'[Node: {self.name}][iter_sum: {self.iter}][Iter: {_iter}/{T}][Batch: {self.n_batch}/{len(self.loader)}][Loss: {loss:.7f} (Avg: {self.train_meter.avg:.7f})]')
+            self.print(f'[iter_sum: {self.iter}][Iter: {_iter}/{T}][Batch: {self.n_batch}/{len(self.loader)}][Loss: {loss:.6f} (Avg: {self.train_meter.avg:.6f})]')
 
             self.loss_tracker.step(self.iter, loss)
 
@@ -319,7 +335,7 @@ class Trainer:
         # When data is given as a dict
         elif datatype is dict:
             data = {key: value.to(self._device) for key, value in data.items()}
-            N = len(next(iter(data))) # number of data in batch
+            N = len(next(iter(data.values()))) # number of data in batch
             outputs = self.forward(**data)
 
         else:
@@ -418,6 +434,16 @@ class Trainer:
             del state_dict['threshold']
 
         self.__dict__.update(state_dict) # update attributes
+        
+    def load_best_model(self):
+        if self.earlystopper is not None:
+            if self.earlystopper.best_network != None:
+                self.print(f'Loading best network: {self.earlystopper.best_network}')
+                self.network.load_state_dict(torch.load(self.earlystopper.best_network))
+            else:
+                self.print(f'No best_network found in self.earlystopper. No network loaded.')
+        else:
+            self.print('self.earlystopper not defined. No network loaded.')
 
     def save(self, path=None, best=True):
         '''
@@ -425,7 +451,7 @@ class Trainer:
         state_dict() -> path/state_dict.p
         network.state_dict() -> path/network.pt
         '''
-        path = self.node_dir if path is None else path
+        path = self.model_dir if path is None else path
         self.print(f'[Node: {self.name}]')
 
         state_dict_path = os.path.join(path, 'node.p')
@@ -456,7 +482,7 @@ class Trainer:
         path/node.p -> state_dict
         path/network.pt -> network.state_dict
         '''
-        path = self.node_dir if path is None else path
+        path = self.model_dir if path is None else path
         self.print(f'[Node: {self.name}]')
 
         state_dict_path = os.path.join(path, 'node.p')
